# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
import io
import logging
from threading import Lock
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import matplotlib
import matplotlib.colors
import numpy as np
from PIL import Image

from .cmap_lc import LAND_COVER_CCI_CMAP
from .cmap_lc import register_lc_color_map


_LOG = logging.getLogger('ect')

CMapTemplateEntry = Tuple[str, str, Sequence[str]]
CMapEntry = Tuple[str, str, Sequence[Tuple[str, str]]]

# Have colormaps separated into categories:
# (taken from http://matplotlib.org/examples/color/colormaps_reference.html)

_CMAP_TEMPLATES: Sequence[CMapTemplateEntry] = (
    ('Perceptually Uniform Sequential',
     'For many applications, a perceptually uniform colormap is the best choice - '
     'one in which equal steps in data are perceived as equal steps in the color space',
     ('viridis', 'inferno', 'plasma', 'magma')),
    ('Sequential 1',
     'These colormaps are approximately monochromatic colormaps varying smoothly '
     'between two color tones - usually from low saturation (e.g. white) to high '
     'saturation (e.g. a bright blue). Sequential colormaps are ideal for '
     'representing most scientific data since they show a clear progression from '
     'low-to-high values.',
     ('Blues', 'BuGn', 'BuPu',
      'GnBu', 'Greens', 'Greys', 'Oranges', 'OrRd',
      'PuBu', 'PuBuGn', 'PuRd', 'Purples', 'RdPu',
      'Reds', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd')),
    ('Sequential 2',
     'Many of the values from the Sequential 2 plots are monotonically increasing.',
     ('afmhot', 'autumn', 'bone', 'cool',
      'copper', 'gist_heat', 'gray', 'hot',
      'pink', 'spring', 'summer', 'winter')),
    ('Diverging',
     'These colormaps have a median value (usually light in color) and vary '
     'smoothly to two different color tones at high and low values. Diverging '
     'colormaps are ideal when your data has a median value that is significant '
     '(e.g.  0, such that positive and negative values are represented by '
     'different colors of the colormap).',
     ('BrBG', 'bwr', 'coolwarm', 'PiYG', 'PRGn', 'PuOr',
      'RdBu', 'RdGy', 'RdYlBu', 'RdYlGn', 'Spectral',
      'seismic')),
    ('Qualitative',
     'These colormaps vary rapidly in color. Qualitative colormaps are useful for '
     'choosing a set of discrete colors.',
     ('Accent', 'Dark2', 'Paired', 'Pastel1',
      'Pastel2', 'Set1', 'Set2', 'Set3')),
    ('Miscellaneous',
     'Colormaps that don\'t fit into the categories above.',
     (LAND_COVER_CCI_CMAP, 'gist_earth', 'terrain', 'ocean', 'gist_stern',
      'brg', 'CMRmap', 'cubehelix',
      'gnuplot', 'gnuplot2', 'gist_ncar',
      'nipy_spectral', 'jet', 'rainbow',
      'gist_rainbow', 'hsv', 'flag', 'prism'))
)

_CMAPS: Optional[Sequence[CMapEntry]] = None

_LOCK = Lock()


def get_cmaps():
    """
    Return a JSON-serializable tuple containing records of the form:
        (<cmap-category>, <cmap-category-description>, <cmap-tuples>),
    where <cmap-tuples> is a tuple containing records
    of the form (<cmap-name>, <cbar-png-bytes>), and where
    <cbar-png-bytes> are encoded PNG images of size 256 x 2 pixels,
    :return: all known matplotlib color maps
    """

    global _CMAPS
    ensure_cmaps_loaded()
    return _CMAPS


def ensure_cmaps_loaded():
    """
    Loads all color maps from matplotlib and registers additional ones,
    if not done before.
    """
    global _CMAP_TEMPLATES, _CMAPS
    if _CMAPS is None:
        _LOCK.acquire()
        if _CMAPS is None:
            register_lc_color_map()
            new_cmaps: List[CMapEntry] = []
            for cmap_category, cmap_description, cmap_names in _CMAP_TEMPLATES:
                cbar_list = []
                for cmap_name in cmap_names:
                    # noinspection PyBroadException
                    try:
                        cmap = matplotlib.colormaps.get_cmap(cmap_name)
                    except Exception:
                        _LOG.warning('detected invalid colormap "%s"' % cmap_name)
                        continue

                    # Add extra colormaps with alpha gradient
                    # see http://matplotlib.org/api/colors_api.html
                    if type(cmap) == matplotlib.colors.LinearSegmentedColormap:
                        new_name = cmap.name + '_alpha'
                        # noinspection PyProtectedMember
                        new_segmentdata = dict(cmap._segmentdata)
                        # let alpha increase from 0.0 to 0.5
                        new_segmentdata['alpha'] = ((0.0, 0.0, 0.0),
                                                    (0.5, 1.0, 1.0),
                                                    (1.0, 1.0, 1.0))
                        new_cmap = matplotlib.colors.LinearSegmentedColormap(
                            new_name, new_segmentdata
                        )
                        matplotlib.colormaps.register(cmap=new_cmap)
                    elif type(cmap) == matplotlib.colors.ListedColormap:
                        new_name = cmap.name + '_alpha'
                        new_colors: List[Sequence[float]] = list(cmap.colors)
                        a_slope = 2.0 / cmap.N
                        a = 0
                        for i in range(len(new_colors)):
                            new_color = new_colors[i]
                            if not isinstance(new_color, str):
                                if len(new_color) == 3:
                                    # noinspection PyTupleAssignmentBalance
                                    r, g, b = new_color
                                    new_colors[i] = r, g, b, a
                                elif len(new_color) == 4:
                                    r, g, b, a_old = new_color
                                    new_colors[i] = r, g, b, min(a, a_old)
                            a += a_slope
                            if a > 1.0:
                                a = 1.0
                        new_cmap = matplotlib.colors.ListedColormap(new_colors, name=new_name)
                        matplotlib.colormaps.register(cmap=new_cmap)
                    else:
                        new_name = cmap.name + '_alpha' \
                            if hasattr(cmap, 'name') else 'unknown'
                        _LOG.warning('could not create colormap "{}" '
                                     'because "{}" is of unknown type {}'
                                     .format(new_name, cmap.name, type(cmap)))

                    gradient = np.linspace(0, 1, 256)
                    gradient = np.vstack((gradient, gradient))
                    image_data = cmap(gradient, bytes=True)
                    image = Image.fromarray(image_data, 'RGBA')

                    ostream = io.BytesIO()
                    image.save(ostream, format='PNG')
                    cbar_png_bytes = ostream.getvalue()
                    ostream.close()

                    cbar_png_data = base64.b64encode(cbar_png_bytes)
                    cbar_png_bytes = cbar_png_data.decode('unicode_escape')

                    cbar_list.append((cmap_name, cbar_png_bytes))
                new_cmaps.append((cmap_category, cmap_description, tuple(cbar_list)))
            _CMAPS = tuple(new_cmaps)
        _LOCK.release()



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Operations for coregistration of datasets

Operations
==========

coregister - coregister two datasets that are defined on pixel-registered grids
that are equidistant in spatial coordinates.

"""
from typing import Tuple

import numpy as np
import xarray as xr
import math

from esa_climate_toolbox.core.op import op_input, op, op_return
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.util.monitor import Monitor

from esa_climate_toolbox.ops import resampling
from esa_climate_toolbox.ops.normalize import adjust_spatial_attrs


@op(tags=['geometric', 'coregistration'],
    version='1.1')
@op_input('method_us', value_set=['nearest', 'linear'])
@op_input(
    'method_ds', value_set=['first', 'last', 'mean', 'mode', 'var', 'std']
)
@op_return(add_history=True)
def coregister(ds_primary: xr.Dataset,
               ds_replica: xr.Dataset,
               method_us: str = 'linear',
               method_ds: str = 'mean',
               monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Perform coregistration of two datasets by resampling the replica dataset
    onto the grid of the primary. If upsampling has to be performed, this is
    achieved using interpolation, if downsampling has to be performed, the
    pixels of the replica dataset are aggregated to form a coarser grid.

    The returned dataset will contain the lat/lon intersection of provided
    primary and replica datasets, resampled unto the primary grid frequency.

    This operation works on datasets whose spatial dimensions are defined on
    pixel-registered grids that are equidistant in lat/lon coordinates, i.e.,
    data points define the middle of a pixel and pixels have the same size
    across the dataset.

    This operation will resample all variables in a dataset, as the lat/lon
    grid is defined per dataset. It works only if all variables in the dataset
    have lat and lon as dimensions.

    For an overview of downsampling/upsampling methods used in this operation,
    please see https://github.com/CAB-LAB/gridtools

    Whether upsampling or downsampling has to be performed is determined
    automatically based on the relationship of the grids of the provided
    datasets.

    :param ds_primary: The dataset whose grid is used for resampling
    :param ds_replica: The dataset that will be resampled
    :param method_us: Interpolation method to use for upsampling.
    :param method_ds: Interpolation method to use for downsampling.
    :param monitor: a progress monitor.
    :return: The replica dataset resampled on the grid of the primary
    """
    try:
        grids = (('replica', ds_replica['lat'].values, -90),
                 ('replica', ds_replica['lon'].values, -180),
                 ('primary', ds_primary['lat'].values, -90),
                 ('primary', ds_primary['lon'].values, -180))
    except KeyError:
        raise ValidationError(
            'Coregistration requires that both datasets are spatial datasets '
            'with lon and lat dimensions. The dimensionality of the provided '
            'primary dataset is: {}, the dimensionality of the provided '
            'replica dataset is: {}. Running the normalize operation might '
            'help in case spatial dimensions have different names'.format(
                ds_primary.dims, ds_replica.dims)
        )

    # Don't do anything if datasets already have the same spatial definition
    if _grids_equal(ds_primary, ds_replica):
        return ds_replica

    # Check if all arrays of the replica dataset have the required
    # dimensionality
    for key in ds_replica.data_vars:
        if not _is_valid_array(ds_replica[key]):
            raise ValidationError(
                '{} data array of replica dataset is not valid for '
                'coregistration. The data array is expected to have lat and '
                'lon dimensions. The data array has the following dimensions: '
                '{}. Consider running select_var operation to exclude this '
                'data array'.format(
                    key, ds_replica[key].dims)
            )

    # Check if the grids of the provided datasets are equidistant and pixel
    # registered
    for array in grids:
        if not _within_bounds(array[1], array[2]):
            raise ValidationError(
                'The {} dataset grid does not fall into required boundaries. '
                'Required boundaries are ({}, {}), dataset boundaries are '
                '({}, {}). Running the normalize operation may help.'.format(
                    array[0], array[2], abs(array[2]), array[1][0], array[1][-1]
                )
            )
        if not _is_equidistant(array[1]):
            raise ValidationError(
                'The {} dataset grid is not equidistant, can not perform '
                'coregistration'.format(array[0])
            )

        if not _is_pixel_registered(array[1], array[2]):
            raise ValidationError(
                'The {} dataset grid is not pixel-registered, can not perform '
                'coregistration'.format(array[0])
            )

    # Co-register
    methods_us = {'nearest': 10, 'linear': 11}
    methods_ds = {
        'first': 50, 'last': 51, 'mean': 54, 'mode': 56, 'var': 57, 'std': 58
    }

    return _resample_dataset(
        ds_primary, ds_replica, methods_us[method_us], methods_ds[method_ds],
        monitor
    )


def _is_equidistant(array: np.ndarray) -> bool:
    """
    Check if the given 1D array is equidistant. The distance between all
    elements of the array should be equal.

    :param array: The array that should be equidistant
    """
    step = abs(array[1] - array[0])
    for i in range(0, len(array) - 1):
        curr_step = abs(array[i + 1] - array[i])
        if not math.isclose(curr_step, step, rel_tol=1e-3):
            return False

    return True


def _is_pixel_registered(array: np.ndarray, origin) -> bool:
    """
    Check if the given coordinate array is pixel registered. Values should
    denote the 'middle' point of a pixel.

    :param array: The array that should be pixel registered
    :param origin: The origin value for the values in the given array
    """
    step = abs(array[1] - array[0])
    return math.isclose(
        (((array[0] - step / 2) - origin) % step), 0, abs_tol=0.1
    )


def _is_valid_array(array: xr.DataArray) -> bool:
    """
    Check if the provided xarray Data Array is valid for coregistration.
    It must have more than two dimensions and at least contain lat and lon.

    :param array: Array to check for validity
    :return: True if the given array is valid
    """
    return (len(array.dims) >= 2
            and 'lat' in array.dims
            and 'lon' in array.dims)


def _within_bounds(array: np.ndarray, low_bound) -> bool:
    """
    Check if the given array falls into the given bounds. We work with
    grids that are symmetrical around zero.

    :param array: Array to check
    :param low_bound: lower boundary
    :return: True if falls into bounds
    """
    return array[0] >= low_bound and array[-1] <= abs(low_bound)


def _resample_slice(
        arr_slice: xr.DataArray, w: int, h: int, ds_method: int, us_method: int,
        parent_monitor: Monitor
) -> xr.DataArray:
    """
    Resample a single time slice of a larger xr.DataArray

    :param arr_slice: xr.DataArray single slice
    :param w: The desired new width (amount of longitudes)
    :param h: The desired new height (amount of latitudes)
    :param ds_method: Downsampling method, see resampling.py
    :param us_method: Upsampling method, see resampling.py
    :param parent_monitor: the parent progress monitor.
    :return: resampled slice
    """
    monitor = parent_monitor.child(1)
    with monitor.observing("resample slice"):
        # In some cases the grouped dimension is not automatically squeezed out
        result = resampling.resample_2d(
            np.ma.masked_invalid(arr_slice.squeeze().values),
            w, h, ds_method, us_method
        )
        return xr.DataArray(result)


def _resample_array(
        array: xr.DataArray, lon: xr.DataArray, lat: xr.DataArray,
        method_us: int, method_ds: int, parent_monitor: Monitor
) -> xr.DataArray:
    """
    Resample the given xr.DataArray to a new grid defined by lat and lon

    :param array: xr.DataArray with lat,lon and time coordinates
    :param lat: 'lat' xr.DataArray attribute for the new grid
    :param lon: 'lon' xr.DataArray attribute for the new grid
    :param method_us: Interpolation method to use for upsampling,
        see resampling.py
    :param method_ds: Interpolation method to use for downsampling,
        see resampling.py
    :param parent_monitor: the parent progress monitor.
    :return: The resampled array
    """
    # Determine width and height of the resampled array
    width = lon.values.size
    height = lat.values.size

    monitor = parent_monitor.child(1)

    kwargs = {
        'w': width, 'h': height, 'ds_method': method_ds,
        'us_method': method_us, 'parent_monitor': monitor
    }

    groupby_list = list(array.dims)
    for dim in ['lon', 'lat']:
        groupby_list.remove(dim)

    if 0 == len(groupby_list):
        # a 2d dataset, can't do groupby => do a simple slice resample
        with monitor.starting("coregister dataarray", total_work=1):
            temp_array = _resample_slice(array, **kwargs)
            coords = {'lat': lat, 'lon': lon}
            return xr.DataArray(temp_array.values,
                                name=array.name,
                                dims=array.dims,
                                coords=coords,
                                attrs=array.attrs).chunk()

    num_steps = 1
    for dim in groupby_list:
        num_steps = num_steps * len(array[dim])

    with monitor.starting("coregister data array", total_work=num_steps):
        temp_array = _nested_groupby_apply(
            array, groupby_list, _resample_slice, kwargs
        )
        chunks = {'lat': height, 'lon': width}
        coords = {'lat': lat, 'lon': lon}
        for dim in groupby_list:
            coords[dim] = array[dim]
            # One spatial slice is one dask chunk, i.e., chunking is
            # (1, 1, 1..1, len(lat), len(lon))
            chunks[dim] = 1
        return xr.DataArray(temp_array.values,
                            name=array.name,
                            dims=array.dims,
                            coords=coords,
                            attrs=array.attrs).chunk(chunks=chunks)


def _resample_dataset(
        ds_primary: xr.Dataset, ds_replica: xr.Dataset, method_us: int,
        method_ds: int, monitor: Monitor
) -> xr.Dataset:
    """
    Resample replica onto the grid of the master.
    This method spatially resamples the whole dataset, e.g., all variables
    in the replica dataset.
    It works only if both datasets have (time, lat, lon) dimensions.

    Note that dataset attributes are not propagated due to currently undecided
    CDM attributes' set.

    :param ds_primary: xr.Dataset whose lat/lon coordinates are used as the
        resampling grid
    :param ds_replica: xr.Dataset that will be resampled on the primary grid
    :param method_us: Interpolation method for upsampling, see resampling.py
    :param method_ds: Interpolation method for downsampling, see resampling.py
    :param monitor: a progress monitor.
    :return: xr.Dataset The resampled replica dataset
    """
    # Find lat/lon bounds of the intersection of master and replica grids. The
    # bounds should fall on pixel boundaries for both spatial dimensions for
    # both datasets
    lat_min, lat_max = _find_intersection(ds_primary['lat'].values,
                                          ds_replica['lat'].values,
                                          global_bounds=(-90, 90))
    lon_min, lon_max = _find_intersection(ds_primary['lon'].values,
                                          ds_replica['lon'].values,
                                          global_bounds=(-180, 180))

    # Subset replica dataset and master grid. We're not using here the subset
    # operation, because the subset operation may produce datasets that cross
    # the anti-meridian by design. However, such a disjoint dataset can not be
    # resampled using our current resampling methods.
    lat_slice = slice(lat_min, lat_max)
    lon_slice = slice(lon_min, lon_max)

    lon = ds_primary['lon'].sel(lon=lon_slice)
    lat = ds_primary['lat'].sel(lat=lat_slice)
    ds_replica = ds_replica.sel(lon=lon_slice, lat=lat_slice)

    # Don't do anything if datasets already have the same spatial definition
    if _grids_equal(ds_primary, ds_replica):
        return ds_replica

    with monitor.starting("coregister dataset", len(ds_replica.data_vars)):
        kwargs = {
            'lon': lon, 'lat': lat, 'method_us': method_us,
            'method_ds': method_ds, 'parent_monitor': monitor
        }
        ret_set = ds_replica.map(_resample_array, keep_attrs=True, **kwargs)

    return adjust_spatial_attrs(ret_set)


def _grids_equal(primary: xr.Dataset,
                 replica: xr.Dataset) -> bool:
    """
    Check if the spatial grid of the given datasets is (almost) equal

    :param primary: Primary dataset
    :param replica: Replica dataset
    """
    if len(primary.lon) != len(replica.lon) or \
       len(primary.lat) != len(replica.lat):
        return False

    if not np.allclose(primary.lon.values, replica.lon.values) or \
       not np.allclose(primary.lat.values, replica.lat.values):
        return False

    return True


def _find_intersection(
        first: np.ndarray, second: np.ndarray,
        global_bounds: Tuple[float, float]
) -> Tuple[float, float]:
    """
    Find 1D intersection of given arrays such that the resulting intersection
    bounds fall on 'pixel' boundaries for both given arrays.

    :param first: First 1D array
    :param second: Second 1D array
    :param global_bounds: (min, max) maximum interval for a valid intersection
    :return: (min, max) intersection bounds
    """
    first_px_size = abs(first[1] - first[0])
    second_px_size = abs(second[1] - second[0])

    minimum = max(first[0] - first_px_size / 2,
                  second[0] - second_px_size / 2)
    maximum = min(first[-1] + first_px_size / 2,
                  second[-1] + second_px_size / 2)

    delta = maximum - minimum
    if delta < max(first_px_size, second_px_size):
        raise ValidationError('Could not find a valid intersection to perform'
                              ' coregistration on')

    # Make sure min/max fall on pixel boundaries for both grids
    # Because there exists a number N denoting how many smaller pixels fall
    # into one larger pixel (for pixel registered datasets with the same
    # origin) => the boundary has to be adjusted by steps equal
    # to smaller pixels.
    finer = min(first_px_size, second_px_size)
    safety = 100
    i = 0
    while (not math.isclose
        (((minimum - global_bounds[0]) % first_px_size), 0, abs_tol=0.1)
           and not math.isclose(
                ((minimum - global_bounds[0]) % second_px_size), 0, abs_tol=0.1)
    ):
        if i == safety:
            raise ValidationError('Could not find a valid intersection to '
                                  'perform coregistration on')
        minimum = minimum + finer
        i = i + 1

    i = 0
    while (not math.isclose(
            ((global_bounds[1] - maximum) % first_px_size), 0, abs_tol=0.1)
           and not math.isclose(
                ((global_bounds[1] - maximum) % second_px_size), 0, abs_tol=0.1
            )
    ):
        if i == safety:
            raise ValidationError('Could not find a valid intersection to '
                                  'perform coregistration on')
        maximum = maximum - finer
        i = i + 1

    # This is possible in some cases when mis-aligned grid arrays are presented
    if maximum <= minimum:
        raise ValidationError('Could not find a valid intersection to perform'
                              ' coregistration on')

    return minimum, maximum


def _nested_groupby_apply(array: xr.DataArray,
                          groupby: list,
                          apply_fn: object,
                          kwargs: dict):
    """
    Perform a nested groupby over given dimensions and apply a function on the
    last 'slice'

    :param array: xr.DataArray to perform groupby on
    :param groupby: a list of coordinate labels over which to perform groupby
    :param apply_fn: The function to apply
    :return: groupby-split-map result
    """
    if len(groupby) == 1:
        return array.groupby(groupby[0], squeeze=False).map(apply_fn, **kwargs)
    else:
        return array.groupby(groupby[0], squeeze=False).map(
            _nested_groupby_apply, groupby=groupby[1:],
            apply_fn=apply_fn, kwargs=kwargs
        )



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from datetime import datetime
from typing import Optional, Sequence, Union, Tuple

import cftime
import numpy as np
import xarray as xr
from matplotlib import path
from shapely.geometry import box, LineString, Polygon

from .types import PolygonLike, ValidationError
from ..util.misc import to_list
from ..util.monitor import Monitor

__author__ = "Janis Gailis (S[&]T Norway)" \
             "Norman Fomferra (Brockmann Consult GmbH)"

DatetimeTypes = np.datetime64, cftime.datetime, datetime
Datetime = Union[np.datetime64, cftime.datetime, datetime]


def adjust_temporal_attrs_impl(ds: xr.Dataset) -> xr.Dataset:
    """
    Adjust the global temporal attributes of the dataset by doing some
    introspection of the dataset and adjusting the appropriate attributes
    accordingly.

    In case the determined attributes do not exist in the dataset, these will
    be added.

    For more information on suggested global attributes see
    `Attribute Convention for Data Discovery
    <http://wiki.esipfed.org/index.php/
    Attribute_Convention_for_Data_Discovery>`_

    :param ds: Dataset to adjust
    :return: Adjusted dataset
    """

    temporal_attrs = _get_temporal_cf_attrs_from_var(ds)

    if temporal_attrs:
        ds = ds.copy()
        # Align temporal attributes with the ones from the shallow Dataset copy
        for key in temporal_attrs:
            if temporal_attrs[key] is not None:
                ds.attrs[key] = temporal_attrs[key]
            else:
                ds.attrs.pop(key, None)

    return ds


def _get_temporal_cf_attrs_from_var(ds: xr.Dataset, var_name: str = 'time') \
        -> Optional[dict]:
    """
    Get temporal boundaries, resolution and duration of the given dataset. If
    the 'bounds' are explicitly defined, these will be used for calculation,
    otherwise it will rest on information gathered from the 'time' dimension
    itself.

    :param ds: Dataset
    :param var_name: The variable/dimension name.
    :return: A dictionary {'attr_name': attr_value}
    """

    if var_name not in ds:
        return None

    var = ds[var_name]

    if 'bounds' in var.attrs:
        # According to CF Conventions the corresponding 'bounds' coordinate
        # variable name should be in the attributes of the coordinate variable
        bnds_name = var.attrs['bounds']
    else:
        # If 'bounds' attribute is missing, the bounds coordinate variable may
        # be named "<dim>_bnds"
        bnds_name = f'{var_name}_bnds'

    dim_min = dim_max = None
    dim_var = None

    if bnds_name in ds:
        bnds_var = ds[bnds_name]
        if len(bnds_var.shape) == 2 and \
                bnds_var.shape[0] > 0 and \
                bnds_var.shape[1] == 2:
            dim_var = bnds_var
            if bnds_var.shape[0] > 1:
                dim_min = bnds_var[0, 0]
                dim_max = bnds_var[-1, 1]
            else:
                dim_min = bnds_var[0, 0]
                dim_max = bnds_var[0, 1]

    if dim_var is None:
        if len(var.shape) == 1 and var.shape[0] > 0:
            dim_var = var
            if var.shape[0] > 1:
                dim_min = var[0]
                dim_max = var[-1]
            else:
                dim_min = var[0]
                dim_max = var[0]

    # Make sure dim_min and dim_max are valid and are time instances
    if dim_var is None \
            or not _is_supported_time_dtype(dim_var.dtype):
        # Cannot determine temporal extent for dimension var_name
        return None

    dim_min = dim_min.values
    dim_max = dim_max.values

    if dim_min != dim_max:
        duration = _get_duration(dim_min, dim_max)
    else:
        duration = None

    if dim_min < dim_max and len(var) >= 2:
        resolution = _get_temporal_res(var)
    else:
        resolution = None

    return dict(time_coverage_start=_to_isoformat(dim_min),
                time_coverage_end=_to_isoformat(dim_max),
                time_coverage_duration=duration,
                time_coverage_resolution=resolution)


def _is_supported_time_dtype(dtype: np.dtype) -> bool:
    return any(np.issubdtype(dtype, t) for t in DatetimeTypes)


def _to_isoformat(time_value: np.ndarray) -> str:
    if np.issubdtype(time_value.dtype, cftime.datetime):
        return time_value.item().isoformat()
    else:
        return np.datetime_as_string(time_value, unit='s')


def _get_temporal_res(time: xr.DataArray) -> str:
    """
    Determine temporal resolution of the given datetimes array.

    See also: `ISO 8601 Durations
    <https://en.wikipedia.org/wiki/ISO_8601#Durations>`_

    :param time: A numpy array containing np.datetime64 objects
    :return: Temporal resolution formatted as an ISO 8601:2004 duration string
    """
    delta = time[1] - time[0]
    days = delta.values.astype('timedelta64[D]') / np.timedelta64(1, 'D')

    if (27 < days) and (days < 32):
        return 'P1M'
    else:
        return 'P{}D'.format(int(days))


def _get_duration(tmin: Datetime, tmax: Datetime) -> str:
    """
    Determine the duration of the given datetimes.

    See also: `ISO 8601 Durations
    <https://en.wikipedia.org/wiki/ISO_8601#Durations>`_

    :param tmin: Time minimum
    :param tmax: Time maximum
    :return: Temporal resolution formatted as an ISO 8601:2004 duration string
    """
    day = np.timedelta64(1, 'D')
    days = (np.timedelta64(tmax - tmin).
            astype(dtype='timedelta64[D]') / day) + 1
    return 'P{}D'.format(int(days))


def get_lon_dim_name_impl(ds: Union[xr.Dataset, xr.DataArray]) -> Optional[str]:
    """
    Get the name of the longitude dimension.
    :param ds: An xarray Dataset
    :return: the name or None
    """
    return _get_dim_name(ds, ['lon', 'longitude', 'long'])


def get_lat_dim_name_impl(ds: Union[xr.Dataset, xr.DataArray]) -> Optional[str]:
    """
    Get the name of the latitude dimension.
    :param ds: An xarray Dataset
    :return: the name or None
    """
    return _get_dim_name(ds, ['lat', 'latitude'])


def _get_dim_name(ds: Union[xr.Dataset, xr.DataArray],
                  possible_names: Sequence[str]) -> Optional[str]:
    for name in possible_names:
        if name in ds.dims:
            return name
    return None


def _lat_inverted(lat: xr.DataArray) -> bool:
    """
    Determine if the latitude is inverted
    """
    if lat[0] > lat[-1]:
        return True

    return False


# TODO (forman): idea: introduce ExtentLike type,
#  or make this a class method of PolygonLike and GeometryLike
def get_extents(
        region: PolygonLike.TYPE
) -> Tuple[Tuple[float, float, float, float], bool]:
    """
    Get extents of a PolygonLike, handles explicitly given
    coordinates.

    :param region: PolygonLike to introspect
    :return: ([lon_min, lat_min, lon_max, lat_max], boolean_explicit_coords)
    """
    explicit_coords = False
    extents = None
    # noinspection PyBroadException
    try:
        maybe_rectangle = to_list(region, dtype=float)
        if maybe_rectangle is not None:
            if len(maybe_rectangle) == 4:
                extents = maybe_rectangle
                explicit_coords = True
    except BaseException:
        # The polygon must be convertible, but it's complex
        polygon = PolygonLike.convert(region)
        if not polygon.is_valid:
            # Heal polygon, see #506 and Shapely User Manual
            region = polygon.buffer(0)
        else:
            region = polygon
        # Get the bounding box
        extents = region.bounds

    if extents:
        return extents, explicit_coords
    else:
        raise ValidationError('Could not determine extents of a polygon')


def _pad_extents(ds: xr.Dataset, extents: Sequence[float]):
    """
    Pad extents by half a pixel in both directions, to make sure subsetting
    slices include all pixels crossed by the bounding box. Set extremes
    to maximum valid geospatial extents.
    """
    try:
        lon_pixel = abs(ds.lon[1] - ds.lon[0])
        lon_min = extents[0] - lon_pixel / 2
        lon_max = extents[2] + lon_pixel / 2
    except IndexError:
        # 1D dimension, leave extents as they were
        lon_min = extents[0]
        lon_max = extents[2]

    try:
        lat_pixel = abs(ds.lat[1] - ds.lat[0])
        lat_min = extents[1] - lat_pixel / 2
        lat_max = extents[3] + lat_pixel / 2
    except IndexError:
        lat_min = extents[1]
        lat_max = extents[3]

    lon_min = -180 if lon_min < -180 else lon_min
    lat_min = -90 if lat_min < -90 else lat_min
    lon_max = 180 if lon_max > 180 else lon_max
    lat_max = 90 if lat_max > 90 else lat_max

    return lon_min, lat_min, lon_max, lat_max


def reset_non_spatial(ds_source: xr.Dataset, ds_target: xr.Dataset):
    """
    Find non spatial data arrays in ds_source and set the corresponding
    data variables in ds_target to original ones.

    :param ds_source: Source dataset
    :param ds_target: Target dataset
    """
    non_spatial = list()
    for var_name in ds_source.data_vars.keys():
        if 'lat' not in ds_source[var_name].dims and \
                'lon' not in ds_source[var_name].dims:
            non_spatial.append(var_name)

    retset = ds_target
    for var in non_spatial:
        retset[var] = ds_source[var]

    return retset


def subset_spatial_impl(ds: xr.Dataset,
                        region: PolygonLike.TYPE,
                        mask: bool = True,
                        monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Do a spatial subset of the dataset

    :param ds: Dataset to subset
    :param region: Spatial region to subset
    :param mask: Should values falling in the bounding box of the polygon but
    not the polygon itself be masked with NaN.
    :param monitor: optional progress monitor
    :return: Subset dataset
    """

    # Validate whether lat and lon exists.

    if not hasattr(ds, 'lon') or not hasattr(ds, 'lat'):
        raise ValidationError(
            'Cannot apply regional subset. No (valid) geocoding found.'
        )

    if hasattr(ds, 'lon') and len(ds.lon.shape) != 1 \
            or hasattr(ds, 'lat') and len(ds.lat.shape) != 1:
        raise ValidationError(
            'Geocoding not recognised. '
            'Variables "lat" and/or "lon" have more than one dimension.'
        )

    monitor.start('Subset', 10)
    # Validate input
    try:
        polygon = PolygonLike.convert(region)
    except BaseException as e:
        raise e

    extents, explicit_coords = get_extents(region)

    lon_min, lat_min, lon_max, lat_max = extents

    # Validate the bounding box
    if (not (-90 <= lat_min <= 90)) or \
            (not (-90 <= lat_max <= 90)) or \
            (not (-180 <= lon_min <= 180)) or \
            (not (-180 <= lon_max <= 180)):
        raise ValidationError(
            'Provided polygon extends outside of geo-spatial bounds. '
            'Latitudes should be from -90 to 90 and '
            'longitudes from -180 to 180 degrees.')

    # Don't do the computationally intensive masking if the provided
    # region is a simple box-polygon, for which there will be nothing to
    # mask.
    simple_polygon = polygon.equals(box(lon_min, lat_min, lon_max, lat_max))

    # Pad extents to include crossed pixels
    lon_min, lat_min, lon_max, lat_max = _pad_extents(ds, extents)

    crosses_antimeridian = (lon_min > lon_max) \
        if explicit_coords \
        else _crosses_antimeridian(polygon)
    lat_inverted = _lat_inverted(ds.lat)
    if lat_inverted:
        lat_index = slice(lat_max, lat_min)
    else:
        lat_index = slice(lat_min, lat_max)

    if crosses_antimeridian and not simple_polygon and mask:
        # Unlikely but plausible
        raise NotImplementedError(
            'Spatial subsets crossing the anti-meridian are currently '
            'implemented for simple, rectangular polygons only, '
            'or complex polygons without masking.')
    monitor.progress(1)
    if crosses_antimeridian:
        # TODO (forman): reimplement entirely, respect dask arrays,
        #  use roll() array operation
        warnings.warn(
            'Entering inefficient code block that requires refactoring. '
            'Expect long execution time and/or out of memory errors.'
        )

        # Shapely messes up longitudes if the polygon crosses the antimeridian
        if not explicit_coords:
            lon_min, lon_max = lon_max, lon_min

        # Can't perform a simple selection with slice, hence we have to
        # construct an appropriate longitude indexer for selection
        lon_left_of_idl = slice(lon_min, 180)
        lon_right_of_idl = slice(-180, lon_max)
        lon_index = xr.concat((ds.lon.sel(lon=lon_right_of_idl),
                               ds.lon.sel(lon=lon_left_of_idl)), dim='lon')

        indexers = {'lon': lon_index, 'lat': lat_index}
        ds_subset = ds.sel(**indexers)

        # Preserve the original longitude dimension, masking elements that
        # do not belong to the polygon with NaN.
        with monitor.observing('subset'):
            return reset_non_spatial(ds, ds_subset.reindex_like(ds.lon))

    lon_slice = slice(lon_min, lon_max)
    indexers = {'lat': lat_index, 'lon': lon_slice}
    ds_subset = ds.sel(**indexers)

    if len(ds_subset.lat) == 0 or len(ds_subset.lon) == 0:
        # Empty return dataset => region out of dataset bounds
        raise ValueError("Can not select a region outside dataset boundaries.")

    if not mask or simple_polygon or explicit_coords:
        # The polygon doesn't cross the anti-meridian,
        # it is a simple box -> Use a simple slice
        with monitor.observing('subset'):
            return reset_non_spatial(ds, ds_subset)

    # TODO (forman): reimplement following code entirely, respect dask arrays,
    #  use efficient mask operations from xcube
    warnings.warn('Entering inefficient code block that requires refactoring.'
                  ' Expect long execution time and/or out of memory errors.')

    # Create the mask array. The result of this is a lon/lat DataArray where
    # all pixels falling in the region or on its boundary are denoted with True
    # and all the rest with False. Works on polygon exterior

    polypath = path.Path(np.column_stack([polygon.exterior.coords.xy[0],
                                          polygon.exterior.coords.xy[1]]))

    # Handle also a single pixel and 1D edge cases
    ds_subset_lon_values = ds_subset.lon.values
    ds_subset__lat_values = ds_subset.lat.values
    if len(ds_subset.lat) == 1 or len(ds_subset.lon) == 1:
        # Create a mask directly on pixel centers
        lonm, latm = np.meshgrid(ds_subset.lon.values, ds_subset.lat.values)
        grid_points = [
            (lon, lat) for lon, lat in zip(lonm.ravel(), latm.ravel())
        ]
        mask_arr = polypath.contains_points(grid_points)
        mask_arr = mask_arr.reshape(lonm.shape)
        mask_arr = xr.DataArray(
            mask_arr,
            coords={'lon': ds_subset.lon, 'lat': ds_subset.lat},
            dims=[ds_subset.lat.dims[0], ds_subset.lon.dims[0]]
        )

        with monitor.observing('subset'):
            # Apply the mask to data
            ds_subset = ds_subset.where(mask_arr, drop=True)

        return reset_non_spatial(ds, ds_subset)

    # The normal case
    # Create a grid of pixel vertices
    lon_pixel = abs(ds.lon[1] - ds.lon[0])
    lat_pixel = abs(ds.lat[1] - ds.lat[0])
    lon_min = float(ds_subset.lon[0] - lon_pixel / 2.)
    lon_max = float(ds_subset.lon[-1] + lon_pixel / 2.)
    lat_min = float(ds_subset.lat[0] - lat_pixel / 2.)
    lat_max = float(ds_subset.lat[-1] + lat_pixel / 2.)

    lat_grid = np.linspace(
        lat_min, lat_max, ds_subset.lat.size + 1, dtype='float32'
    )
    lon_grid = np.linspace(
        lon_min, lon_max, ds_subset.lon.size + 1, dtype='float32'
    )

    try:
        lonm, latm = np.meshgrid(lon_grid, lat_grid)
    except MemoryError:
        raise ValidationError(
            'Not enough memory to mask the dataset with the given polygon. '
            'Try subsetting with masking disabled'
        )

    # Mark all grid points falling within the polygon as True

    monitor.progress(1)
    try:
        grid_points = np.empty(
            (len(lat_grid) * len(lon_grid), 2), dtype='float32'
        )
    except MemoryError:
        raise ValidationError(
            'Not enough memory to mask the dataset with the given polygon. '
            'Try subsetting with masking disabled.'
        )
    for i, el in enumerate(((lon, lat)
                            for lon, lat in zip(lonm.ravel(), latm.ravel()))):
        grid_points[i] = el

    monitor.progress(1)
    try:
        mask_arr = polypath.contains_points(grid_points)
    except MemoryError:
        raise ValidationError(
            'Not enough memory to mask the dataset with the given polygon. '
            'Try subsetting with masking disabled.'
        )

    monitor.progress(1)

    mask_arr = mask_arr.reshape(lonm.shape)
    # Vectorized 'rolling window' numpy magic to go
    # from pixel vertices to pixel centers
    mask_arr = mask_arr[1:, 1:] + mask_arr[1:, :-1] + \
               mask_arr[:-1, 1:] + mask_arr[:-1, :-1]

    mask_arr = xr.DataArray(
        mask_arr,
        coords={'lon': ds_subset_lon_values, 'lat': ds_subset__lat_values},
        dims=['lat', 'lon']
    )

    with monitor.observing('subset'):
        # Apply the mask to data
        ds_subset = ds_subset.where(mask_arr, drop=False)

    monitor.done()
    return reset_non_spatial(ds, ds_subset)


def _crosses_antimeridian(region: Polygon) -> bool:
    """
    Determine if the given region crosses the Antimeridian line, by converting
    the given Polygon from -180;180 to 0;360 and checking if the antimeridian
    line crosses it.

    This only works with Polygons without holes

    :param region: Polygon to test
    """

    # Convert region to only have positive longitudes.
    # This way we can perform a simple antimeridian check

    old_exterior = region.exterior.coords
    new_exterior = []
    for point in old_exterior:
        lon, lat = point[0], point[1]
        if -180. <= lon < 0.:
            lon += 360.
        new_exterior.append((lon, lat))
    converted_region = Polygon(new_exterior)

    # There's a problem at this point. Any polygon crossed by the zero-th
    # meridian can in principle convert to an inverted polygon that is crossed
    # by the antimeridian.

    if not converted_region.is_valid:
        # The polygon 'became' invalid upon conversion => probably the original
        # polygon is what we want

        # noinspection PyBroadException
        try:
            # First try cleaning up geometry that is invalid
            converted_region = converted_region.buffer(0)
        except BaseException:
            pass
        if not converted_region.is_valid:
            return False

    test_line = LineString([(180, -90), (180, 90)])
    if test_line.crosses(converted_region):
        # The converted polygon seems to be valid and crossed by the
        # antimeridian. At this point there's no 'perfect' way how to tell if
        # we wanted the converted polygon or the original one.

        # A simple heuristic is to check for size. The smaller one is quite
        # likely the desired one
        if converted_region.area < region.area:
            return True
        else:
            return False
    else:
        return False


def subset_temporal_impl(ds: xr.Dataset,
                         time_range: Tuple[datetime, datetime]) -> xr.Dataset:
    """
    Do a temporal subset of the dataset.

    :param ds: Dataset or dataframe to subset
    :param time_range: Time range to select
    :return: Subset dataset
    """
    # If it can be selected, go ahead
    try:
        # noinspection PyTypeChecker
        time_slice = slice(time_range[0], time_range[1])
        indexers = {'time': time_slice}
        return ds.sel(**indexers)
    except TypeError:
        raise ValueError(
            'Time subset operation expects a dataset with the time coordinate '
            'of type datetime64[ns], but received {}. Running the normalize '
            'operation on this dataset may help'.format(ds.time.dtype))


def subset_temporal_index_impl(ds: xr.Dataset,
                               time_ind_min: int,
                               time_ind_max: int) -> xr.Dataset:
    """
    Do a temporal indices based subset

    :param ds: Dataset or dataframe to subset
    :param time_ind_min: Minimum time index to select
    :param time_ind_max: Maximum time index to select
    :return: Subset dataset
    """
    # we're creating a slice that includes both ends
    # to have the same functionality as subset_temporal
    time_slice = slice(time_ind_min, time_ind_max + 1)
    indexers = {'time': time_slice}
    return ds.isel(**indexers)



from unittest import TestCase

from esa_climate_toolbox.core import plugin


class PluginTest(TestCase):

    def test_that_test_plugins_are_loaded(self):
        self.assertIsNotNone(plugin.PLUGIN_REGISTRY)

    def test_ect_init(self):
        # Yes, this is really a silly test :)
        # But this way we cover one more (empty) statement.
        plugin.ect_init(True, False, a=1, b=2)



"""
Tests for animation operations
"""
import itertools
import os
import sys
import shutil
import tempfile
from contextlib import contextmanager
from unittest import TestCase
import unittest

import xarray as xr
import numpy as np
import pandas as pd

from esa_climate_toolbox.ops.animate import animate_map

_counter = itertools.count()
ON_WIN = sys.platform == 'win32'


@contextmanager
def create_tmp_file(name, ext):
    """
    Create a temporary filename for testing
    """
    tmp_dir = tempfile.mkdtemp()
    path = os.path.join(tmp_dir, '{}{}.{}'.format(name, next(_counter), ext))
    try:
        yield path
    finally:
        try:
            shutil.rmtree(tmp_dir)
        except OSError:
            if not ON_WIN:
                raise


@unittest.skipIf(condition=os.environ.get('ECT_DISABLE_PLOT_TESTS', None),
                 reason="skipped if ECT_DISABLE_PLOT_TESTS=1")
class TestAnimateMap(TestCase):
    """
    Test animate_map() operation
    """

    def test_animate_map(self):
        # Test the nominal functionality. This doesn't check that the animation
        # is what is expected, rather, it simply tests if it seems to have been created
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'second': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'time': pd.date_range('2000-01-01', periods=2)})

        with create_tmp_file('remove_me', 'html') as tmp_file:
            animate_map(dataset, file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test if extents can be used
        with create_tmp_file('remove_me', 'html') as tmp_file:
            animate_map(dataset,
                        var='second',
                        region='-40.0, -20.0, 50.0, 60.0',
                        file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test setting a Title
        with create_tmp_file('remove_me', 'html') as tmp_file:
            animate_map(dataset,
                        var='second',
                        title='Title',
                        file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test using a global color map
        with create_tmp_file('remove_me', 'html') as tmp_file:
            animate_map(dataset,
                        var='second',
                        true_range=True,
                        file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

    def test_plot_map_exceptions(self):
        # Test if the corner cases are detected without creating a plot for it.

        # Test value error is raised when passing an unexpected dataset type
        with create_tmp_file('remove_me', 'html') as tmp_file:
            with self.assertRaises(ValueError) as cm:
                animate_map([1, 2, 4], file=tmp_file)
            self.assertEqual(str(cm.exception),
                             "Input 'ds' for operation "
                             "'esa_climate_toolbox.ops.animate.animate_map' "
                             "must be of type 'Dataset', but got type 'list'.")
            self.assertFalse(os.path.isfile(tmp_file))

        # Test the extensions bound checking
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.random.rand(2, 4, 1)),
            'lat': np.linspace(-89.5, 89.5, 2),
            'lon': np.linspace(-179.5, 179.5, 4)})

        with self.assertRaises(ValueError):
            region = '-40.0, -95.0, 50.0, 60.0',
            animate_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-40.0, -20.0, 50.0, 95.0',
            animate_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-181.0, -20.0, 50.0, 60.0',
            animate_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-40.0, -20.0, 181.0, 60.0',
            animate_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-40.0, -20.0, 50.0, -25.0',
            animate_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-20.0, -20.0, -25.0, 60.0',
            animate_map(dataset, region=region)

        # Test all-nan dataset
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.full([5, 10, 2], np.nan)),
            'second': (['lat', 'lon', 'time'], np.full([5, 10, 2], np.nan)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'time': pd.date_range('2000-01-01', periods=2)})

        with create_tmp_file('remove_me', 'html') as tmp_file:
            with self.assertRaises(ValueError):
                animate_map(dataset,
                            var='second',
                            true_range=True,
                            file=tmp_file)



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Anomaly calculation operations

Functions
=========
"""
import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.util.monitor import Monitor
from esa_climate_toolbox.ops.subset import subset_spatial
from esa_climate_toolbox.ops.subset import subset_temporal
from esa_climate_toolbox.ops.arithmetics import diff
from esa_climate_toolbox.ops.arithmetics import arithmetics
from esa_climate_toolbox.core.types import PolygonLike
from esa_climate_toolbox.core.types import TimeRangeLike
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.ops.normalize import adjust_spatial_attrs
from esa_climate_toolbox.ops.normalize import adjust_temporal_attrs


_ALL_FILE_FILTER = dict(name='All Files', extensions=['*'])


@op(tags=['anomaly'], version='1.1')
@op_input('file', file_open_mode='r',
          file_filters=[dict(name='NetCDF', extensions=['nc']),
                        _ALL_FILE_FILTER])
@op_return(add_history=True)
def anomaly_external(ds: xr.Dataset,
                     file: str,
                     transform: str = None,
                     monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Calculate anomaly with external reference data, for example, a climatology.
    The given reference dataset is expected to consist of 12 time slices, one
    for each month.

    The returned dataset will contain the variable names found in both - the
    reference and the given dataset. Names found in the given dataset, but not in
    the reference, will be dropped from the resulting dataset. The calculated
    anomaly will be against the corresponding month of the reference data.
    E.g. January against January, etc.

    In case spatial extents differ between the reference and the given dataset,
    the anomaly will be calculated on the intersection.

    :param ds: The dataset to calculate anomalies from
    :param file: Path to reference data file
    :param transform: Apply the given transformation before calculating the anomaly.
                      For supported operations see help on 'ds_arithmetics' operation.
    :param monitor: a progress monitor.
    :return: The anomaly dataset
    """
    # Check if the time coordinate is of dtype datetime
    try:
        if ds.time.dtype != 'datetime64[ns]':
            raise ValidationError('The dataset provided for anomaly calculation'
                                  ' is required to have a time coordinate of'
                                  ' dtype datetime64[ns]. Running the normalize'
                                  ' operation on this dataset might help.')
    except AttributeError:
        raise ValidationError('The dataset provided for anomaly calculation'
                              ' is required to have a time coordinate.')

    try:
        if ds.attrs['time_coverage_resolution'] != 'P1M':
            raise ValidationError(
                'anomaly_external expects a monthly dataset, '
                'got: {} instead.'.format(ds.attrs['time_coverage_resolution']))
    except KeyError:
        try:
            ds = adjust_temporal_attrs(ds)
            if ds.attrs['time_coverage_resolution'] != 'P1M':
                raise ValidationError(
                    'anomaly_external expects a monthly dataset, '
                    'got: {} instead.'.format(ds.attrs['time_coverage_resolution']))
        except KeyError:
            raise ValidationError('Could not determine temporal resolution of'
                                  ' of the given input dataset.')

    clim = xr.open_dataset(file)
    try:
        if len(clim.time) != 12:
            raise ValidationError(
                'The reference dataset is expected to be a monthly climatology. '
                'The provided dataset has a time dimension with length: {}'.format(
                    len(clim.time))
            )
    except AttributeError:
        raise ValidationError('The reference dataset is required to '
                              'have a time coordinate.')

    ret = ds.copy()
    if transform:
        ret = arithmetics(ds, transform)
    # Group by months, subtract the appropriate slice from the reference
    # Note that this requires that 'time' coordinate labels are of type
    # datetime64[ns]
    total_work = 100
    step = 100 / 12

    with monitor.starting('Anomaly', total_work=total_work):
        monitor.progress(work=0)
        kwargs = {'ref': clim, 'monitor': monitor, 'step': step}
        ret = ret.groupby(ds['time.month']).map(_group_anomaly, **kwargs)

    # Running groupby results in a redundant 'month' variable being added to
    # the dataset (dask < 2021.0)
    if "month" in ret:
        ret = ret.drop_vars(['month'])

    ret.attrs = ds.attrs
    # The dataset may be cropped
    return adjust_spatial_attrs(ret)


def _group_anomaly(group: xr.Dataset,
                   ref: xr.Dataset,
                   monitor: Monitor = Monitor.NONE,
                   step: float = None):
    """
    Calculate anomaly for the given group.

    :param group: Result of a groupby('time.month') operation
    :param ref: Reference dataset
    :param monitor: Monitor of the parent method
    :param step: Step to add to monitor progress
    :return: Group dataset with anomaly calculation applied
    """
    # Retrieve the month of the current group
    month = group['time.month'][0].values
    ret = diff(group, ref.isel(time=month - 1))
    monitor.progress(work=step)
    return ret


@op(tags=['anomaly'], version='1.0')
@op_input('time_range', data_type=TimeRangeLike)
@op_input('region', data_type=PolygonLike)
@op_return(add_history=True)
def anomaly_internal(ds: xr.Dataset,
                     time_range: TimeRangeLike.TYPE = None,
                     region: PolygonLike.TYPE = None,
                     monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Calculate anomaly using as reference data the mean of an optional region
    and time slice from the given dataset. If no time slice/spatial region is
    given, the operation will calculate anomaly using the mean of the whole
    dataset as the reference.

    This is done for each data array in the dataset.
    :param ds: The dataset to calculate anomalies from
    :param time_range: Time range to use for reference data
    :param region: Spatial region to use for reference data
    :param monitor: a progress monitor.
    :return: The anomaly dataset
    """
    ref = ds.copy()
    if time_range:
        time_range = TimeRangeLike.convert(time_range)
        ref = subset_temporal(ref, time_range)
    if region:
        region = PolygonLike.convert(region)
        ref = subset_spatial(ref, region)
    with monitor.observing("Calculating anomaly"):
        ref = ref.mean(keep_attrs=True, skipna=True)
        anomaly_diff = ds - ref
    return anomaly_diff



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
from typing import Dict, Any, Callable

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"


def _get_safe_globals_accessor() -> Callable[[], dict]:
    safe_builtin_names = [
        "abs",
        "all",
        "any",
        "ascii",
        "bin",
        "bool",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "complex",
        "dict",
        "divmod",
        "enumerate",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "hasattr",
        "hash",
        "hex",
        "int",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "list",
        "map",
        "max",
        "min",
        "next",
        "object",
        "oct",
        "ord",
        "pow",
        "range",
        "repr",
        "reversed",
        "round",
        "set",
        "slice",
        "sorted",
        "str",
        "sum",
        "tuple",
        "type",
        "zip",
    ]

    safe_module_names = [
        'cmath',
        'math',
    ]

    builtins = sys.modules['builtins']

    safe_globals = {}
    for name in safe_builtin_names:
        safe_globals[name] = builtins.__dict__[name]

    for name in safe_module_names:
        safe_globals[name] = __import__(name)

    safe_globals['__builtins__'] = None
    safe_globals['builtins'] = None

    def _get_safe_globals_closure():
        """Return a global environment for safe expression evaluation."""
        return dict(safe_globals)

    return _get_safe_globals_closure


get_safe_globals = _get_safe_globals_accessor()


def safe_eval(expression: str, local_namespace: Dict[str, Any] = None):
    """
    Evaluate the given Python *expression* in the given *local_namespace*.

    The **expression** argument is parsed and evaluated as a Python expression
    using the **local_namespace** mapping as local namespace.
    By default, the **expression** has no access to the current environment and
    only limited access to the standard builtins, i.e., only functions
    considered safe are allowed, e.g. *abs*, *min*, *max*, etc.

    Syntax errors are reported as exceptions.

    :param expression: A Python expression.
    :param local_namespace: The local namespace in which **expression**
        is evaluated.
    :return: The result of the evaluated expression.
    """
    return eval(expression, get_safe_globals(), local_namespace or {})


def safe_exec(source_code: str, local_namespace: Dict[str, Any] = None):
    """
    Execute the given *source_code* in the in the given *local_namespace*.

    The **source_code** argument is parsed and evaluated as a Python code
    using the **local_namespace** mapping as local namespace.
    By default, the **source_code** has no access to the current environment
    and only limited access to the standard builtins, i.e. only functions
    considered safe are allowed, e.g. *abs*, *min*, *max*, etc.

    Syntax errors are reported as exceptions.

    :param source_code: Python source code.
    :param local_namespace: The local namespace in which **expression**
        is evaluated.
    :return: The result of the evaluated expression.
    """
    return exec(source_code, get_safe_globals(), local_namespace or {})



"""
Tests for outlier detection operations
"""

from unittest import TestCase
import xarray as xr
import numpy as np

from esa_climate_toolbox.ops import outliers


class TestOutliers(TestCase):
    def test_outliers(self):
        ds = xr.Dataset({
            'first': xr.DataArray(np.arange(16, dtype=float).reshape(4, 4),
                                  dims=('x', 'y'),
                                  attrs={'a1': 'Dummy attribute'}),
            'second': xr.DataArray(np.arange(16).reshape(4, 4),
                                   dims=('x', 'y'),
                                   attrs={'a1': 'Dummy attribute'})
        }, attrs={'dummy': 'dummy_attr'})

        # Test nominal execution, outliers -> nan
        ret_ds = outliers.detect_outliers(ds, 'first')
        test = ds['first'].copy()
        test[0][0] = np.nan
        test[3][3] = np.nan
        self.assertTrue(test.identical(ret_ds['first']))
        self.assertTrue(ret_ds['second'].identical(ds['second']))

        # Test nominal execution with a wildcard
        ret_ds = outliers.detect_outliers(ds, 'fir*', threshold_low=2,
                                          threshold_high=13, quantiles=False)
        test = ds['first'].copy()
        test[0] = [np.nan, np.nan, np.nan, 3.0]
        test[3] = [12.0, np.nan, np.nan, np.nan]
        self.assertTrue(test.identical(ret_ds['first']))
        self.assertTrue(ret_ds['second'].identical(ds['second']))

        # Test adding a mask data array to a dataset that doesn't abide
        # strictly by CF conventions
        ret_ds = outliers.detect_outliers(ds, 'first', mask=True)
        self.assertTrue(ds['first'].equals(ret_ds['first']))
        test_mask = np.array(np.zeros(16, dtype='i1').reshape(4, 4))
        test_mask[0][0] = 1
        test_mask[3][3] = 1
        ret_mask = ret_ds['first_outlier_mask']
        ret_first = ret_ds['first']
        self.assertTrue(np.array_equal(test_mask, ret_mask.values))
        self.assertTrue('status_flag' in ret_mask.attrs['standard_name'])
        self.assertTrue('mask' in ret_mask.attrs['long_name'])
        self.assertTrue(('first_outlier_mask' in
                         ret_first.attrs['ancillary_variables']))

        # Test adding a mask data array to a dataset that somewhat abides by CF
        # conventions
        ds = xr.Dataset({
            'first': xr.DataArray(np.arange(16, dtype=float).reshape(4, 4),
                                  dims=('x', 'y'),
                                  attrs={'a1': 'Dummy attribute',
                                         'ancillary_variables': 'second',
                                         'standard_name': 'dummy',
                                         'long_name': 'Dummy'}),
            'second': xr.DataArray(np.arange(16).reshape(4, 4),
                                   dims=('x', 'y'),
                                   attrs={'a1': 'Dummy attribute'})
        }, attrs={'dummy': 'dummy_attr'})

        ret_ds = outliers.detect_outliers(ds, 'first', mask=True)
        self.assertTrue(ds['first'].equals(ret_ds['first']))
        test_mask = np.array(np.zeros(16, dtype='i1').reshape(4, 4))
        test_mask[0][0] = 1
        test_mask[3][3] = 1
        ret_mask = ret_ds['first_outlier_mask']
        ret_first = ret_ds['first']
        self.assertTrue(np.array_equal(test_mask, ret_mask.values))
        self.assertTrue('status_flag' in ret_mask.attrs['standard_name'])
        self.assertTrue('dummy ' in ret_mask.attrs['standard_name'])
        self.assertTrue('mask' in ret_mask.attrs['long_name'])
        self.assertTrue('Dummy ' in ret_mask.attrs['long_name'])
        self.assertTrue(('first_outlier_mask' in
                         ret_first.attrs['ancillary_variables']))
        self.assertTrue(('second ' in
                         ret_first.attrs['ancillary_variables']))



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Subset operations

Components
==========
"""
from typing import Dict

import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.core.opimpl import subset_spatial_impl
from esa_climate_toolbox.core.opimpl import subset_temporal_impl
from esa_climate_toolbox.core.opimpl import subset_temporal_index_impl
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import DictLike
from esa_climate_toolbox.core.types import PolygonLike
from esa_climate_toolbox.core.types import PointLike
from esa_climate_toolbox.core.types import TimeRangeLike
from esa_climate_toolbox.ops.normalize import adjust_spatial_attrs
from esa_climate_toolbox.ops.normalize import adjust_temporal_attrs
from esa_climate_toolbox.util.misc import to_scalar
from esa_climate_toolbox.util.monitor import Monitor
from esa_climate_toolbox.util.undefined import UNDEFINED

from xcube.core.normalize import get_geo_spatial_attrs_from_var


@op(tags=['geometric', 'spatial', 'subset'], version='1.0')
@op_input('region', data_type=PolygonLike)
@op_return(add_history=True)
def subset_spatial(ds: xr.Dataset,
                   region: PolygonLike.TYPE,
                   mask: bool = True,
                   monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Do a spatial subset of the dataset

    :param ds: Dataset to subset
    :param region: Spatial region to subset
    :param mask: Should values falling in the bounding box of the polygon but not the
        polygon itself be masked with NaN.
    :param monitor: A monitor to report the progress of the process
    :return: Subset dataset
    """
    return adjust_spatial_attrs(
        subset_spatial_impl(ds, region, mask, monitor), allow_point=True
    )


@op(tags=['subset', 'temporal', 'filter'], version='1.0')
@op_input('ds', data_type=DatasetLike)
@op_input('time_range', data_type=TimeRangeLike)
@op_return(add_history=True)
def subset_temporal(ds: DatasetLike.TYPE,
                    time_range: TimeRangeLike.TYPE) -> xr.Dataset:
    """
    Do a temporal subset of the dataset.

    :param ds: Dataset or dataframe to subset
    :param time_range: Time range to select
    :return: Subset dataset
    """
    ds = DatasetLike.convert(ds)
    time_range = TimeRangeLike.convert(time_range)
    return adjust_temporal_attrs(subset_temporal_impl(ds, time_range))


@op(tags=['subset', 'temporal', 'filter', 'utility'], version='1.0')
@op_input('ds', data_type=DatasetLike)
@op_return(add_history=True)
def subset_temporal_index(ds: DatasetLike.TYPE,
                          time_ind_min: int,
                          time_ind_max: int) -> xr.Dataset:
    """
    Do a temporal indices based subset

    :param ds: Dataset or dataframe to subset
    :param time_ind_min: Minimum time index to select
    :param time_ind_max: Maximum time index to select
    :return: Subset dataset
    """
    ds = DatasetLike.convert(ds)
    return subset_temporal_index_impl(ds, time_ind_min, time_ind_max)


def extract_point(ds: DatasetLike.TYPE,
                  point: PointLike.TYPE,
                  indexers: DictLike.TYPE = None,
                  tolerance_default: float = 0.01) -> Dict:
    """
    Extract data at the given point location. The returned dict will contain scalar
    values for all variables for which all dimension have been given in ``indexers``.
    For the dimensions *lon* and *lat* a nearest neighbour lookup is performed.
    All other dimensions must mach exact.

    :param ds: Dataset or dataframe to subset
    :param point: Geographic point given by longitude and latitude
    :param indexers: Optional indexers into data array of *var*. The *indexers* is a
        dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels. e.g. "layer=4".
    :param tolerance_default: The default longitude and latitude tolerance for the
        nearest neighbour lookup.
        It will only be used, if it is not possible to deduce the resolution of the
        dataset.
    :return: A dict with the scalar values of all variables and the variable names as
        keys.
    """
    ds = DatasetLike.convert(ds)
    point = PointLike.convert(point)
    indexers = DictLike.convert(indexers) or {}

    lon_lat_indexers = {'lon': point.x, 'lat': point.y}
    tolerance = _get_tolerance(ds, tolerance_default)

    variable_values = {}
    var_names = sorted(ds.data_vars.keys())
    for var_name in var_names:
        if not var_name.endswith('_bnds'):
            variable = ds.data_vars[var_name]
            effective_indexers = {}
            used_dims = {'lat', 'lon'}
            for dim_name, dim_value in indexers.items():
                if dim_name in variable.dims:
                    effective_indexers[dim_name] = dim_value
                    used_dims.add(dim_name)
            if set(variable.dims) == used_dims:
                try:
                    lon_lat_data = variable.sel(**effective_indexers)
                except KeyError:
                    # if there is no exact match for the "additional" dims,
                    # skip this variable
                    continue
                try:
                    point_data = lon_lat_data.sel(method='nearest',
                                                  tolerance=tolerance,
                                                  **lon_lat_indexers)
                except KeyError:
                    # if there is no point within the given tolerance,
                    # return an empty dict
                    return {}
                if not variable_values:
                    variable_values['lat'] = float(point_data.lat)
                    variable_values['lon'] = float(point_data.lon)
                value = to_scalar(point_data.values, num_digits=3)
                if value is not UNDEFINED:
                    variable_values[var_name] = value
    return variable_values


def _get_tolerance(ds: xr.Dataset, tolerance_default: float):
    lon_res_attr_name = 'geospatial_lon_resolution'
    lat_res_attr_name = 'geospatial_lat_resolution'
    if lat_res_attr_name in ds.attrs and lon_res_attr_name in ds.attrs:
        lon_res = ds.attrs[lon_res_attr_name]
        lat_res = ds.attrs[lat_res_attr_name]
    else:
        try:
            lon_res = get_geo_spatial_attrs_from_var(ds, 'lon')[lon_res_attr_name]
            lat_res = get_geo_spatial_attrs_from_var(ds, 'lat')[lat_res_attr_name]
        except ValueError:
            return tolerance_default
    if isinstance(lon_res, str) and lon_res.find(' ') > 0:
        lon_res = lon_res.split(' ')[0]
    if isinstance(lat_res, str) and lat_res.find(' ') > 0:
        lat_res = lat_res.split(' ')[0]
    try:
        return (float(lon_res) + float(lat_res)) / 2
    except ValueError:
        return tolerance_default



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Dataset normalization operation.

Components
==========
"""

import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.core.opimpl import adjust_temporal_attrs_impl
import xcube.core.normalize as xcube_normalize


@op(tags=['utility'], version='1.0')
@op_return(add_history=True)
def normalize(ds: xr.Dataset) -> xr.Dataset:
    """
    Normalize the geo- and time-coding upon opening the given dataset w.r.t.
    to a common (CF-compatible) convention used within the ESA Climate Toolbox.
    This will maximize the compatibility of a dataset for usage with operations.

    That is,
    * variables named "latitude" will be renamed to "lat";
    * variables named "longitude" or "long" will be renamed to "lon";

    Then, for equi-rectangular grids,
    * Remove 2D "lat" and "lon" variables;
    * Two new 1D coordinate variables "lat" and "lon" will be generated from
    original 2D forms.

    Finally, it will be ensured that a "time" coordinate variable will be of
    type *datetime*.

    :param ds: The dataset to normalize.
    :return: The normalized dataset, or the original dataset,
        if it is already "normal".
    """
    return xcube_normalize.normalize_dataset(ds, reverse_decreasing_lat=True)


@op(tags=['utility'], version='1.0')
def adjust_spatial_attrs(
        ds: xr.Dataset, allow_point: bool = False
) -> xr.Dataset:
    """
    Adjust the global spatial attributes of the dataset by doing some
    introspection of the dataset and adjusting the appropriate attributes
    accordingly.

    In case the determined attributes do not exist in the dataset, these will
    be added.

    For more information on suggested global attributes see
    `Attribute Convention for Data Discovery
    <http://wiki.esipfed.org/index.php/Attribute_Convention_for_Data_Discovery>`_

    :param ds: Dataset to adjust
    :param allow_point: Whether a dataset containing a single point is allowed
    :return: Adjusted dataset
    """
    return xcube_normalize.adjust_spatial_attrs(ds, allow_point=allow_point)


@op(tags=['utility'], version='1.0')
def adjust_temporal_attrs(ds: xr.Dataset) -> xr.Dataset:
    """
    Adjust the global temporal attributes of the dataset by doing some
    introspection of the dataset and adjusting the appropriate attributes
    accordingly.

    In case the determined attributes do not exist in the dataset, these will
    be added.

    If the temporal attributes exist, but the dataset lacks a variable 'time', a
    new dimension 'time' of size one will be added and related coordinate
    variables 'time' and 'time_bnds' are added to the dataset. The dimension of
    all non-coordinate variables will be expanded by the new time dimension.

    For more information on suggested global attributes see
    `Attribute Convention for Data Discovery
    <http://wiki.esipfed.org/index.php/Attribute_Convention_for_Data_Discovery>`_

    :param ds: Dataset to adjust
    :return: Adjusted dataset
    """
    return adjust_temporal_attrs_impl(ds)



import sys
import time

args = sys.argv

n = 10
if len(args) > 1:
    n = int(args[1])

period = 0.1
if len(args) > 2:
    period = float(args[2])

fail_at = -1
if len(args) > 3:
    fail_at = int(args[3])

print('mkentropy: Running {} steps'.format(n))
for i in range(n):
    if i == fail_at:
        raise RuntimeError('An intended error occurred!')
    time.sleep(period)
    print('mkentropy: Did {} of {} steps: {}%'.format(
        i + 1, n, (100 * (i + 1)) / n)
    )

print('mkentropy: Done making some entropy')



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

The ``esa_climate_toolbox.core.plugin`` module exposes the Toolbox plugin
``REGISTRY`` which is mapping from the toolbox entry point names to plugin meta
information. A Toolbox plugin is any callable in an internal/extension module
registered with ``esa_climate_toolbox_plugins`` entry point.

Clients register a Toolbox plugin in the ``setup()`` call of their ``setup.py``
script. The following plugin example comprises a main module
``wavelet_gapfill`` which provides the entry point function ``ect_init``:::

    setup(
        name="gapfill-wavelet",
        version="0.5",
        description='A wavelet-based gap-filling algorithm',
        license='GPL 3',
        author='John Doe',
        packages=['wavelet_gapfill'],
        entry_points={
            'esa_climate_toolbox_plugins': [
                'wavelet_gapfill = wavelet_gapfill:ect_init',
            ],
        },
        install_requires=['pywavelets >= 2.1'],
    )

The entry point callable should have the following signature::

    def ect_init(*args, **kwargs):
        pass

or::

    class EctInit:
        def __init__(*args, **kwargs)__:
            pass

The return values are ignored.

Verification
============

The module's unit-tests are located in
`test/cire/test_plugin.py <https://github.com/esa-cci/esa-climate-toolbox/tree/main/test/core/test_plugin.py>`_
and may be executed using
``$ py.test test/core/test_plugin.py --cov=esa_climate_toolbox/core/plugin.py``
for extra code coverage information.

Components
==========
"""

import logging
from collections import OrderedDict

from importlib.metadata import entry_points

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

_LOG = logging.getLogger('ect')


def _load_plugins():
    plugins = OrderedDict()
    for entry_point in entry_points().select(
            group='esa_climate_toolbox_plugins'
    ):

        # noinspection PyBroadException
        try:
            plugin = entry_point.load()
        except Exception:
            _LOG.exception(
                "unexpected exception while loading ESA Climate Toolbox plugin "
                "with entry point '%s'" % entry_point.name)
            continue

        if callable(plugin):
            # noinspection PyBroadException
            try:
                plugin()
            except Exception:
                _LOG.exception(
                    "unexpected exception while executing ESA Climate Toolbox "
                    "plugin with entry point '%s'" % entry_point.name)
                continue
        else:
            _LOG.error("ESA Climate Toolbox plugin with entry point '%s' "
                       "must be a callable but got a '%s'" %
                       (entry_point.name, type(plugin)))
            continue

        # Here: use pkg_resources and introspection to generate a
        # JSON-serializable dictionary of plugin meta-information
        plugins[entry_point.name] = {'entry_point': entry_point.name}

    return plugins


def ect_init(*arg, **kwargs):
    """
    No actual use, just demonstrates the signature of an entry point callable.

    :param arg: any arguments (not used)
    :param kwargs: any keyword arguments (not used)
    :return: any or void (not used)
    """
    return arg, kwargs


#: Mapping of ESA Climate Toolbox entry point names to JSON-serializable plugin
# meta-information.
PLUGIN_REGISTRY = _load_plugins()



################################################################################
# This is an ESA Climate Toolbox configuration file.                           #
#                                                                              #
# If this file is "~/.ect/conf.py", it is the active configuration.            #
#                                                                              #
# If this file is named "conf.py.template" you can rename it and move          #
# it to "~/.ect/conf.py" to make it the active configuration file.             #
#                                                                              #
# As this is a regular Python script, you may use any Python code to compute   #
# the settings provided here.                                                  #
#                                                                              #
################################################################################

# 'data_stores_path' is denotes a directory where the ESA Climate Toolbox
# stores information about data stores and also saves local data files
# synchronized with their remote versions.
# Use the tilde '~' (also on Windows) within the path to point to your home
# directory.
#
# data_stores_path = '~/.ect/data_stores'

# Default color map to be used for any variable not configured in
# 'variable_display_settings'
# 'default_color_map' must be the name of a color map taken from from
# https://matplotlib.org/examples/color/colormaps_reference.html
# default_color_map = 'jet'
default_color_map = 'inferno'

# Data Store Configurations
# Load from here the configurations of the data stores that will eventually be
# loaded into the ESA Climate Toolbox
store_configs = {
    "local": {
        "store_id": "file",
        "store_params": {
            "root": "",
        }
     },
     "cci-store": {
        "store_id": "esa-cdc"
     },
     "cci-zarr-store": {
        "store_id": "esa-cdc-zarr"
    }
}



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The core API of the ESA Climate Toolbox.
"""

from .ds import add_local_store
from .ds import add_store
from .ds import find_data_store
from .ds import get_output_store_id
from .ds import get_store
from .ds import get_search_params
from .ds import get_supported_formats
from .ds import list_datasets
from .ds import list_ecvs
from .ds import list_ecv_datasets
from .ds import list_stores
from .ds import open_data
from .ds import remove_store
from .ds import search
from .ds import set_output_store
from .ds import write_data

# noinspection PyUnresolvedReferences
from .op import list_operations
from .op import get_op
from .op import get_op_meta_info
from .op import new_expression_op
from .op import new_subprocess_op
from .op import op
from .op import op_input
from .op import op_output
from .op import op_return
from .op import Operation
from .op import OP_REGISTRY

# noinspection PyUnresolvedReferences
from ..util.monitor import Monitor

# noinspection PyUnresolvedReferences
from ..util.monitor import ChildMonitor

# noinspection PyUnresolvedReferences
from ..util.opmetainf import OpMetaInfo

# Run plugin registration by importing the plugin module
# noinspection PyUnresolvedReferences
from .plugin import ect_init as _

del _



import importlib.resources
from typing import Union, List

import numpy as np
import pandas as pd
from shapely.ops import unary_union
import xarray as xr

from xcube.core.geom import mask_dataset_by_geometry
from xcube.core.geom import rasterize_features
from xcube.core.store import new_data_store
from xcube.util.assertions import assert_instance


def _get_countries_df():
    with importlib.resources.path(
            "esa_climate_toolbox.functions.country_data", "README.md"
    ) as p:
        countries_path = str(p.parent)
    countries_store = new_data_store("file", root=countries_path)
    return countries_store.open_data(f"countries.geojson")


def make_regions_dataset(
        template: xr.Dataset, regions: Union[str, List[str]] = None
) -> xr.Dataset:
    """Rasterize country and continent polygons into a grid provided
    by a *template* dataset.

    The returned dataset has the same gridmapping and same
    spatial chunking as *template* and contains two variables:

    * "country_code" provides an integer country code.
      The names of countries are provided by its attribute "country_names".
    * "continent_code" provides an integer continent code.
      The names of continent are provided by its attribute "continent_names".

    :param template: The dataset to which the regions dataset shall refer
    :param regions: A single name of a country or continent or a list of names.
        If given, only these regions will be considered. If omitted, all countries
        will be included. Default is None.
    :return: A dataset with the same grid mapping as *template*
    """
    assert_instance(template, xr.Dataset, "template")

    if regions is not None and not isinstance(regions, List):
        regions = list(regions)

    countries = _get_countries_df()
    if regions:
        countries_subset = countries[countries['name'].isin(regions)]
        continents_subset = countries[countries['continent'].isin(regions)]
        countries = pd.concat([countries_subset, continents_subset])
        countries = countries.reset_index(drop=True)
    country_names = {code + 1: name for code, name in enumerate(countries['name'])}
    continent_names = {code + 1: name for code, name in
                       enumerate(set(countries['continent']))}

    countries["country_code"] = [code for code in country_names.keys()]
    continent_codes = []
    for continent_name in countries["continent"]:
        for c_code, c_name in continent_names.items():
            if continent_name == c_name:
                continent_codes.append(c_code)
                break
    countries["continent_code"] = continent_codes

    countries_and_continents_dataset = rasterize_features(
        template,
        countries,
        ["country_code", "continent_code"],
        var_props={
            "country_code": {
                "name": "country_code",
                "dtype": np.uint8,
                "fill_value": 255
            },
            "continent_code": {
                "name": "continent_code",
                "dtype": np.uint8,
                "fill_value": 255
            }
        }
    )
    countries_and_continents_dataset = (
        countries_and_continents_dataset.drop_vars("time"))
    country_code = countries_and_continents_dataset.country_code
    country_code.attrs["country_names"] = country_names
    continent_code = countries_and_continents_dataset.continent_code
    continent_code.attrs["continent_names"] = continent_names

    return countries_and_continents_dataset


def get_land_mask(dataset: xr.Dataset):
    """Gets a land mask for a dataset.

    The returned dataset has the same grid mapping and same
    spatial chunking as *template* and contains one variable:

    * "land" provides a mask that is True if the pixel is on land, otherwise False.

    :param dataset: The dataset for which the mask shall be created
    :return: A dataset with the same grid mapping as *dataset*
    """
    countries_dataset = make_regions_dataset(dataset)
    countries_dataset['land'] = xr.where(
        countries_dataset.country_code >= 0, True, False
    )
    countries_dataset = countries_dataset.drop_vars("country_code")
    countries_dataset = countries_dataset.drop_vars("continent_code")
    return countries_dataset


def get_regions_mask(dataset: xr.Dataset, regions: Union[str, List[str]]):
    """Gets a regions mask for a dataset.

    The returned dataset has the same gridmapping and same
    spatial chunking as *template* and contains one variable:

    * "regions" provides a mask that is True if the pixel is in one of the specified
     regions, otherwise False.

    :param dataset: The dataset for which the mask shall be created
    :param regions: A single name of a country or continent or a list of names.
    :return: A dataset with the same grid mapping as *dataset*
    """
    countries_dataset = make_regions_dataset(dataset, regions)

    countries_dataset['regions'] = xr.where(
        countries_dataset.country_code + countries_dataset.continent_code >= 0,
        True, False
    )
    countries_dataset = countries_dataset.drop_vars("country_code")
    countries_dataset = countries_dataset.drop_vars("continent_code")
    return countries_dataset


def mask_dataset_by_land(dataset: xr.Dataset):
    """Masks out non-land-pixels of a dataset.

    :param dataset: The dataset for which the mask shall be created
    :return: A dataset which only has values for pixels on land
    """
    countries = _get_countries_df()
    land_geometry = unary_union(countries.geometry)
    return mask_dataset_by_geometry(dataset, land_geometry)


def mask_dataset_by_regions(dataset: xr.Dataset, regions: Union[str, List[str]]):
    """Masks out pixels of a dataset which are not in one of the specified regions.

    :param dataset: The dataset for which the mask shall be created
    :param regions: A single name of a country or continent or a list of names.
    :return: A dataset which only has values for pixels in one of the specified regions.
    """
    countries = _get_countries_df()

    countries_subset = countries[countries['name'].isin(regions)]
    continents_subset = countries[countries['continent'].isin(regions)]
    sub_df = pd.concat([countries_subset, continents_subset])

    sub_geometry = unary_union(sub_df.geometry)
    return mask_dataset_by_geometry(dataset, sub_geometry)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import fnmatch
import os
import os.path
import re
import sys
import urllib.parse
from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime, date, timedelta
from io import StringIO
from typing import Union, Tuple, Sequence, Optional, Iterable, Any

import numpy as np

from .sround import sround
from .undefined import UNDEFINED

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"


def qualified_name_to_object(qualified_name: str,
                             default_module_name='builtins'):
    """
    Convert a fully qualified name into a Python object.
    It is true that
    ``qualified_name_to_object(object_to_qualified_name(obj)) is obj``.

    >>> qualified_name_to_object('unittest.TestCase')
    <class 'unittest.case.TestCase'>

    See also :py:func:`object_to_qualified_name`.

    :param qualified_name: fully qualified name of the form
        [<module>'.'{<name>'.'}]<name>
    :param default_module_name: default module name to be used if the name
        does not contain one
    :return: the Python object
    :raise ImportError: If the module could not be imported
    :raise AttributeError: If the name could not be found
    """
    parts = qualified_name.split('.')

    if len(parts) == 1:
        module_name = default_module_name
    else:
        module_name = parts[0]
        parts = parts[1:]

    value = __import__(module_name)
    for name in parts:
        value = getattr(value, name)
    return value


def object_to_qualified_name(
        value, fail=False, default_module_name='builtins'
) -> Union[str, None]:
    """
    Get the fully qualified name of a Python object.
    It is true that
    ``qualified_name_to_object(object_to_qualified_name(obj)) is obj``.

    >>> from unittest import TestCase
    >>> object_to_qualified_name(TestCase)
    'unittest.case.TestCase'

    See also :py:func:`qualified_name_to_object`.

    :param value: some Python object
    :param fail: raise ``ValueError`` if name cannot be derived.
    :param default_module_name: if this is the *value*'s module name,
        no module name will be returned.
    :return: fully qualified name if it can be derived,
        ``None`` if *fail* is ``False``.
    :raise ValueError: if *fail* is ``True`` and the name cannot be derived.
    """

    try:
        module_name = value.__module__
    except AttributeError:
        module_name = None
    if module_name == default_module_name:
        module_name = None

    try:
        name = value.__name__
    except AttributeError:
        name = None

    if name:
        return module_name + '.' + name if module_name else name
    elif fail:
        raise ValueError("missing attribute '__name__'")
    else:
        return str(value)


@contextmanager
def fetch_std_streams():
    """
    A context manager which can be used to temporarily fetch the standard
    output streams ``sys.stdout`` and  ``sys.stderr``.

    Usage:::

        with fetch_std_streams() as stdout, stderr
            sys.stdout.write('yes')
            sys.stderr.write('oh no')
        print('fetched', stdout.getvalue())
        print('fetched', stderr.getvalue())

    :return: yields  ``sys.stdout`` and  ``sys.stderr``
    redirected into buffers of type ``StringIO``
    """
    sys.stdout.flush()
    sys.stderr.flush()

    old_stdout = sys.stdout
    old_stderr = sys.stderr

    sys.stdout = StringIO()
    sys.stderr = StringIO()

    try:
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout.flush()
        sys.stderr.flush()

        sys.stdout = old_stdout
        sys.stderr = old_stderr


def encode_url_path(
        path_pattern: str, path_args: dict = None, query_args: dict = None
) -> str:
    """
    Return an URL path with an optional query string which is composed of
    a *path_pattern* that may contain placeholders of the form ``{name}``
    which will be replaced by URL-encoded versions of the corresponding values
    in *path_args*, i.e. ``urllib.parse.quote_plus(path_args[name])``.
    An optional query string is composed of the URL-encoded key-value pairs
    given in *query_args*, i.e. ``urllib.parse.urlencode(query_args)``.

    :param path_pattern: The path pattern which may include any number of
        placeholders of the form ``{name}``
    :param path_args: The values for the placeholders in *path_pattern*
    :param query_args: The query arguments
    :return: an URL-encoded path
    """
    path = path_pattern
    if path_args:
        quoted_pattern_args = dict(path_args)
        for name, value in path_args.items():
            quoted_pattern_args[name] = urllib.parse.quote_plus(str(value)) \
                if value is not None else ''
        path = path_pattern.format(**quoted_pattern_args)
    query_string = ''
    if query_args:
        query_string = '?' + urllib.parse.urlencode(query_args)
    return path + query_string


def to_datetime_range(start_datetime_or_str: Union[datetime, date, str, None],
                      end_datetime_or_str: Union[datetime, date, str, None],
                      default=None) -> Tuple[datetime, datetime]:
    if not start_datetime_or_str and not end_datetime_or_str:
        return default
    if not end_datetime_or_str:
        if not start_datetime_or_str:
            raise ValueError('start_datetime_or_str argument must be given')
        end_datetime_or_str = start_datetime_or_str
    start_datetime = to_datetime(start_datetime_or_str, upper_bound=False)
    end_datetime = to_datetime(end_datetime_or_str, upper_bound=True)
    return start_datetime, end_datetime


def to_datetime(
        datetime_or_str: Union[datetime, date, str, None],
        upper_bound=False, default=None
) -> datetime:
    if datetime_or_str is None:
        return default
    elif isinstance(datetime_or_str, str):
        if datetime_or_str.strip() == '':
            return default
        format_to_timedelta = [("%Y-%m-%dT%H:%M:%S", timedelta()),
                               ("%Y-%m-%d %H:%M:%S", timedelta()),
                               ("%Y-%m-%d", timedelta(hours=24, seconds=-1)),
                               ("%Y-%m", timedelta(weeks=4, seconds=-1)),
                               ("%Y", timedelta(days=365, seconds=-1)),
                               ]
        for f, td in format_to_timedelta:
            try:
                dt = datetime.strptime(datetime_or_str, f)
                return dt + td if upper_bound else dt
            except ValueError:
                pass
        raise ValueError('Invalid date/time value: "%s"' % datetime_or_str)
    elif isinstance(datetime_or_str, datetime):
        return datetime_or_str
    elif isinstance(datetime_or_str, date):
        return datetime(datetime_or_str.year,
                        datetime_or_str.month,
                        datetime_or_str.day,
                        12)
    else:
        raise TypeError('datetime_or_str argument must be a string '
                        'or instance of datetime.date')


def to_list(value,
            dtype: type = str,
            name: str = None,
            nullable: bool = True,
            csv: bool = True,
            strip: bool = True):
    """
    Convert *value* into a list of items of type *dtype*.

    :param value: Some value that may be a sequence or a scalar
    :param dtype: The desired target type
    :param name: An (argument) name used for ``ValueError`` messages
    :param nullable: Whether *value* can be None
    :param csv: Whether to split *value* if it is a string containing commas
    :param strip: Whether to strip CSV string values, used only if *csv* is True
    :return: A list with elements of type *dtype* or
        None if *value* is None and *nullable* is True
    """
    if value is None:
        if not nullable:
            raise ValueError('%s argument must not be None' % (name or 'some'))
        return value
    if csv and isinstance(value, str):
        items = value.split(',')
        return [dtype(item.strip() if strip else item) for item in items]
    if isinstance(value, dtype):
        return [value]
    # noinspection PyBroadException
    try:
        return [dtype(item) for item in value]
    except Exception:
        return [dtype(value)]


_PYTHON_QUOTE_CHARS = ['"', "'"]


def to_str_constant(s: str, quote="'") -> str:
    """
    Convert a given string into another string
        that is a valid Python representation of a string constant.
    :param s: the string
    :param quote: the quote character, either a single or double quote
    :return:
    """
    if s is None:
        raise ValueError()
    if quote not in _PYTHON_QUOTE_CHARS:
        raise ValueError()
    return quote + s.replace('\\', '\\\\').replace(quote,
                                                   "\\%s" % quote) + quote


def is_str_constant(s: str) -> bool:
    """
    Test whether a given string is a Python representation of a string constant.

    :param s: the string
    :return: True, if so.
    """
    return s and len(s) >= 2 and s[0] == s[-1] and s[0] in _PYTHON_QUOTE_CHARS


@contextmanager
def cwd(path: str):
    """
    A context manager which can be used to temporarily change
    the current working directory to *path*.

    Usage:::

        print(os.getcwd())
        with cwd('./test'):
            print(os.getcwd())
        print(os.getcwd())

    :return: yields the new working directory (absolute *path* passed in)
    """
    if path is None:
        raise ValueError('path argument must be given')

    old_dir = os.getcwd()

    try:
        os.chdir(path)
        yield os.getcwd()
    finally:
        os.chdir(old_dir)


_DATETIME64 = np.dtype('datetime64')
_ZERO_THMS_POSTFIX = 'T00:00:00'
_ZERO_MICR_POSTFIX = '.000000000'


def date_to_simple_str(v):
    time_str = str(v)
    if time_str.endswith(_ZERO_MICR_POSTFIX):
        time_str = time_str[0: -len(_ZERO_MICR_POSTFIX)]
    if time_str.endswith(_ZERO_THMS_POSTFIX):
        time_str = time_str[0: -len(_ZERO_THMS_POSTFIX)]
    return time_str


def to_json(v):
    if v is None:
        return v

    t = type(v)
    if t in {bool, int, float, str}:
        return v
    if t == complex:
        return [v.real, v.imag]
    if isinstance(v, type):
        return object_to_qualified_name(v)

    # TODO (forman): handle dtype=uint64/int64 here,
    # as JSON does not support 64-bit ints

    is_datetime64 = False
    try:
        is_datetime64 = np.issubdtype(v.dtype, np.datetime64)
    except AttributeError:
        pass

    if is_datetime64:
        # Convert time values to time strings
        is_scalar = False
        try:
            is_scalar = v.size == 1 and len(v.shape) == 0
        except AttributeError:
            pass
        if is_scalar:
            return date_to_simple_str(v)
        else:
            li = []
            for vi in v:
                li.append(date_to_simple_str(vi))
            return li

    if isinstance(v, np.ndarray) and not np.issubdtype(v.dtype, np.datetime64):
        try:
            return v.tolist()
        except AttributeError:
            pass

    try:
        return v.item()
    except (AttributeError, ValueError):
        pass

    try:
        d = OrderedDict()
        for ki, vi in v.items():
            d[str(ki)] = to_json(vi)
        return d
    except AttributeError:
        pass

    try:
        li = []
        for vi in v:
            li.append(to_json(vi))
        return li
    except TypeError:
        pass

    return str(v)


def filter_fileset(names: Sequence[str],
                   includes: Optional[Sequence[str]] = None,
                   excludes: Optional[Sequence[str]] = None) -> Sequence[str]:
    """
    Filter a fileset given by the sequence *names* using the wildcard patterns
        in *includes* and *excludes*.
    :param names: The names of the fileset
    :param includes: Wildcard patterns that select the file names to be included
    :param excludes: Wildcard patterns that select the file names to be excluded
    :return: The filtered fileset
    """
    if includes is not None:
        filtered_names = set()
        for pattern in includes:
            filtered_names.update(fnmatch.filter(names, pattern))
        if excludes is not None:
            filtered_names_old = filtered_names
            filtered_names = set(filtered_names)
            for pattern in excludes:
                filtered_names.difference_update(
                    fnmatch.filter(filtered_names_old, pattern)
                )
    elif excludes is not None:
        filtered_names = set(names)
        for pattern in excludes:
            filtered_names.difference_update(
                fnmatch.filter(filtered_names, pattern)
            )
    else:
        filtered_names = names
    return filtered_names


def new_indexed_name(names: Iterable[str], pattern: str) -> str:
    """
    Return a new name that is unique in *names* and that conforms to *pattern*.
    The argument *pattern* must contain a single ``"{index}"`` substring.

    :param names: Sequence of names
    :param pattern: Naming pattern, e.g. "var_{index}"
    :return: a new name, e.g. "var_3"
    """
    if "{index}" not in pattern:
        raise ValueError('pattern must contain "{index}"')
    re_pattern = re.compile(pattern.replace("{index}", r"(\d+)"))
    max_index = 0
    for name in names:
        match_result = re_pattern.match(name)
        if match_result and match_result.group(1):
            max_index = max(max_index, int(match_result.group(1)))

    new_index = max_index + 1
    while True:
        new_name = pattern.replace("{index}", str(new_index))
        if not new_name.isidentifier():
            raise ValueError('pattern does not yield a valid name')
        if new_name not in names:
            return new_name
        new_index += 1


NoneType = type(None)


# noinspection PyBroadException
def to_scalar(
        value: Any, num_chars=None, num_digits=None, stringify=False
) -> Any:
    """
    Convert the given *value* into a JSON-serializable, scalar value.
    If the conversion fails, UNDEFINED is returned.

    :param value: Any value.
    :param num_chars: If not None and greater zero,
        text values will be limited to *num_chars* characters
    :param num_digits: If not None, floating point values will be rounded
        to *num_digits* significant digits
    :param stringify: If True, non-primitive values will be converted to strings
    :return: A JSON-serializable, scalar value or UNDEFINED,
        if the conversion fails.
    """
    is_float = False
    is_str = False
    if isinstance(value, (int, bool, NoneType)):
        return value
    elif isinstance(value, float):
        is_float = True
    elif isinstance(value, str):
        is_str = True
    elif hasattr(value, 'shape') and hasattr(value, 'dtype'):
        try:
            shape = value.shape
            dtype = 'object' if str(value.dtype) == 'geometry' else value.dtype
            ndim = len(shape)
            size = 1
            for dim in shape:
                size *= dim
            if size > 1:
                return UNDEFINED
            if ndim >= 1:
                index = 0 if ndim == 1 else (0,) * ndim
                try:
                    value = value[index]
                except BaseException:
                    pass
            if np.issubdtype(dtype, np.integer):
                return int(value)
            elif np.issubdtype(dtype, np.bool_):
                return bool(value)
            elif np.issubdtype(dtype, np.floating):
                value = float(value)
                is_float = True
            elif np.issubdtype(dtype, np.str_) or stringify:
                value = str(value)
                is_str = True
            else:
                return UNDEFINED
        except BaseException as e:
            print("Error in to_scalar: " + str(e))
            return UNDEFINED
    elif stringify:
        value = str(value)
        is_str = True
    else:
        return UNDEFINED

    if is_float:
        if num_digits is not None:
            return sround(value, num_digits=num_digits)
        return value

    if is_str:
        if num_chars is not None and len(value) > num_chars:
            return value[0: num_chars] + '...'
        return value

    return UNDEFINED


_DEBUG_MODE = None


def is_debug_mode() -> bool:
    global _DEBUG_MODE
    if _DEBUG_MODE is None:
        # noinspection PyBroadException
        try:
            _DEBUG_MODE = bool(int(os.getenv('ECT_DEBUG', '0')))
        except Exception:
            _DEBUG_MODE = False
    return _DEBUG_MODE



from unittest import TestCase

from esa_climate_toolbox.util.undefined import UNDEFINED


class UndefinedTest(TestCase):

    def test_undefined(self):
        self.assertIsNotNone(UNDEFINED)
        self.assertEqual(str(UNDEFINED), 'UNDEFINED')
        self.assertEqual(repr(UNDEFINED), 'UNDEFINED')



# The MIT License (MIT)
# Copyright (c) 2016, 2017 by the ESA CCI Toolbox development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


"""
Description
===========

The ``cate.util.im`` package provides application-independent utility functions for working with tiled image pyramids.

The Cate project uses this package for implementing a RESTful web service that provides image tiles from image pyramids.

This package is independent of other ``cate.*`` packages, but it depends on the following external packages

* numpy
* pillow (for PIL)
* matplotlib


Verification
============

The module's unit-tests are located in
`test/util/im <https://github.com/CCI-Tools/cate/blob/master/test/util/im>`_ and may be executed using
``$ py.test test/util/im --cov=cate/util/im`` for extra code coverage information.

Components
==========
"""

from .cmaps import ensure_cmaps_loaded
from .cmaps import get_cmaps

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"



from unittest import TestCase

import xarray as xr

from esa_climate_toolbox.ops.select import select_var


class TestSelect(TestCase):

    def test_select(self):
        dataset = xr.Dataset({'abc': ('x', [1, 2, 3]),
                              'bde': ('x', [4, 5, 6])})

        # Test if nothing gets dropped if nothing has to be dropped
        actual = select_var(dataset)
        self.assertDatasetEqual(dataset, actual)

        actual = select_var(dataset, var='')
        self.assertDatasetEqual(dataset, actual)

        # Test that everything is dropped if the desired name does not exist in
        # the dataset
        expected = xr.Dataset({'abc': ('x', [1, 2, 3])})
        expected = expected.drop_vars(['abc', ])
        actual = select_var(dataset, var='xyz')
        self.assertDatasetEqual(expected, actual)

        # Test that a single variable selection works
        actual = select_var(dataset, var='abc')
        expected = xr.Dataset({'abc': ('x', [1, 2, 3])})
        self.assertDatasetEqual(expected, actual)

        # Test that simple multiple variable selection works
        actual = select_var(dataset, var='abc,bde')
        self.assertDatasetEqual(dataset, actual)

        # Test that wildcard selection works
        actual = select_var(dataset, var='*b*')
        self.assertDatasetEqual(dataset, actual)

    def assertDatasetEqual(self, expected, actual):
        # this method is functionally equivalent to `assert expected == actual`,
        # but it checks each aspect of equality separately for easier debugging
        assert expected.equals(actual), (expected, actual)



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import matplotlib


LAND_COVER_CCI_CMAP = 'land_cover_cci'


def register_lc_color_map():
    lc_color_mappings = [
        (0, dict(r=0, g=0, b=0)),
        (10, dict(r=255, g=255, b=100)),
        (11, dict(r=255, g=255, b=100)),
        (12, dict(r=255, g=255, b=0)),
        (20, dict(r=170, g=240, b=240)),
        (30, dict(r=220, g=240, b=100)),
        (40, dict(r=200, g=200, b=100)),
        (50, dict(r=0, g=100, b=0)),
        (60, dict(r=0, g=160, b=0)),
        (61, dict(r=0, g=160, b=0)),
        (62, dict(r=170, g=200, b=0)),
        (70, dict(r=0, g=60, b=0)),
        (71, dict(r=0, g=60, b=0)),
        (72, dict(r=0, g=80, b=0)),
        (80, dict(r=40, g=80, b=0)),
        (81, dict(r=40, g=80, b=0)),
        (82, dict(r=40, g=100, b=0)),
        (90, dict(r=120, g=130, b=0)),
        (100, dict(r=140, g=160, b=0)),
        (110, dict(r=190, g=150, b=0)),
        (120, dict(r=150, g=100, b=0)),
        (121, dict(r=120, g=75, b=0)),
        (122, dict(r=150, g=100, b=0)),
        (130, dict(r=255, g=180, b=50)),
        (140, dict(r=255, g=220, b=210)),
        (150, dict(r=255, g=235, b=175)),
        (151, dict(r=255, g=205, b=120)),
        (152, dict(r=255, g=210, b=120)),
        (153, dict(r=255, g=235, b=175)),
        (160, dict(r=0, g=120, b=190)),
        (170, dict(r=0, g=150, b=120)),
        (180, dict(r=0, g=220, b=130)),
        (190, dict(r=195, g=20, b=0)),
        (200, dict(r=255, g=245, b=215)),
        (201, dict(r=220, g=220, b=220)),
        (202, dict(r=255, g=245, b=215)),
        (210, dict(r=0, g=70, b=200)),
        (220, dict(r=255, g=255, b=255)),
    ]

    classes = {lc: color for lc, color in lc_color_mappings}

    invalid_rgba = (0, 0, 0, 0.5)
    class_0_rgba = (0, 0, 0, 0)

    rgba_list = []
    num_entries = 256
    last_rgba = invalid_rgba
    for i in range(num_entries):
        color = classes.get(i)
        if color:
            last_rgba = (color['r'] / 255, color['g'] / 255, color['b'] / 255, 1.0)
        rgba_list.append(last_rgba)
    rgba_list[0] = class_0_rgba

    cmap = matplotlib.colors.ListedColormap(
        rgba_list, name=LAND_COVER_CCI_CMAP, N=num_entries
    )
    matplotlib.colormaps.register(cmap=cmap)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import setup, find_packages

requirements = [
    # Sync with ./environment.yml.
]

packages = find_packages(exclude=["test", "test.*"])

# Same effect as "from esa_climate_toolbox import version",
# but avoids importing esa_climate_toolbox:
__version__ = None
with open('esa_climate_toolbox/version.py') as f:
    exec(f.read())

setup(
    name="esa_climate_toolbox",
    version=__version__,
    description='A toolbox that provides access to data from the ESA Open '
                'Data Portal as well as functions to work with this data.',
    license='MIT',
    author='CCI Toolbox Development Team',
    packages=packages,
    include_package_data=True,
    data_files=[(
        'esa_climate_toolbox', [
            'esa_climate_toolbox/ds/data/excluded_data_sources',
            'esa_climate_toolbox/ds/data/dataset_states.json',
            'esa_climate_toolbox/ds/data/stores.yml',
            'esa_climate_toolbox/ds/data/kc_refs.json'
            ]
    )],
    install_requires=requirements,
    entry_points={
        'xcube_plugins': [
            'esa-climate-toolbox = esa_climate_toolbox.plugin:init_plugin',
        ],
        'esa_climate_toolbox_plugins': [
            'esa_climate_toolbox_ops = esa_climate_toolbox.ops:ect_init',
            'esa_climate_toolbox_ds = esa_climate_toolbox.ds:ect_init',
        ],
    }
)
